module NDFSMez where

import Data.Set (Set)
import Data.Char

import qualified Data.Set as Set

{- Program that allows building regular expressions.
   The value of a regular expression is a DFSM that
   accepts only the value described by the regular expression.
   For example star (union (once 'a', once 'b'))
   represents (a U b)*.  Its value is a DFSM that accepts
   all strings involving only a's and b's.
   Use the regular expression in
      match regExp input
   where input is the list of inputs.  It returns true iff input
   is in the meaning of the regular expression.
-}

{- We're being lazy and using ints for states
   This will make it easier to change names when we
   combine ndfsm's.
-}

type State = Int

-- Simple NFSM

-- This definition using records automatically provides
-- functions returning the components!  Cool abbreviations!

-- Element of type Regex i represents a ndfsm & has form
--      Regex numStates startStates transitionFcn acceptStates
-- i is the type of the inputs to the ndfsm, typically Chars 
data Regex i = Regex
    { numberOfStates     :: Int
    , startingStates     :: Set State
    , transitionFunction :: i -> State -> Set State
    , acceptingStates    :: Set State
    } 

shift :: Int -> Set State -> Set State
-- shift states up by n: converts set to ascending list,
-- then increases each elt by n, then converts back to set.
shift n = Set.fromAscList . map (+ n) . Set.toAscList

-- regexp that accepts only x
once :: Eq i => i -> Regex i
once x = Regex n ss tf as
      where
         n = 2
         ss = Set.singleton 0
         as = Set.singleton 1

         tf i 0 | i == x = Set.singleton 1
         tf _ _ = Set.empty
         
-- regex that accepts any single input
dot :: Regex i
dot = Regex n ss tf as
      where
         n = 2
         ss = Set.singleton 0
         as = Set.singleton 1

         tf i 0 = Set.singleton 1
         tf _ _ = Set.empty

-- regex that rejects everything
empty = Regex n ss tf as
      where
        n  = 0
        ss = Set.empty
        as = Set.empty
        tf _ _ = Set.empty

    -- "ε": the regular expression that matches the empty string
epsilon = Regex n ss tf as
      where
        n  = 1
        ss = Set.singleton 0
        as = Set.singleton 0
        tf _ _ = Set.empty

-- produce regular expression that is the union of the regular
-- expressions corresponding to the fst and snd args
union (Regex nL ssL tfL asL) (Regex nR ssR tfR asR) = Regex n ss tf as
      where
        n  = nL + nR
        ss = Set.union ssL (shift nL ssR)
        as = Set.union asL (shift nL asR)
        tf i s | s < nL    = tfL i s
               | otherwise = shift nL (tfR i (s - nL))

-- produce regular expression that is the concatenation of the
-- regular expressions corresponding to the fst and snd args
cat (Regex nL ssL tfL asL) (Regex nR ssR tfR asR) = Regex n ss tf as
      where
        n = nL + nR

        ss =
            if Set.null (Set.intersection ssL asL)
               -- no starting state of left is also accepting state
            then           ssL
            else Set.union ssL (shift nL ssR)

        tf i s =
            if s < nL -- use tfL to transition or move to start of R
            then if Set.null (Set.intersection r asL)
                 then           r
                 -- if r is accepting state of L, allow to move
                 -- to start state of R
                 else Set.union r (shift nL ssR)
            else shift nL (tfR i (s - nL))
            -- if s from right machine, use tfR to transition
          where
            r = tfL i s

        as = shift nL asR  -- only accept states are from r

-- Kleene star operator: regular expression corresponding to zero or
-- more occurrences of regular expression argument.  CHECK!!!!
star :: Regex i -> Regex i
star (Regex n ss tf as) = Regex (n+1) ss' tf' ss'
  where
    tf' i s =
        let r = tf i s
        in  if Set.null (Set.intersection r as)
              -- new state r not an accept state, so move normally
            then           r
              -- new state r is accept state so allow to move to
              -- start state as well. 
            else Set.union r ss
    ss' = Set.union ss (Set.singleton n)
    
-- Plus operator produces regular expression corresponding to
-- one or more occurrences of regular expression argument
plus :: Regex i -> Regex i
plus (Regex n ss tf as) = Regex n ss tf' as
  where
    tf' i s =
        let r = tf i s
        in  if Set.null (Set.intersection r as)
            then           r
            else Set.union r ss

-- takes a regular expression and a list of inputs and determines whether
-- it is accepted or not
match :: Regex i -> [i] -> Bool
match (Regex n ss _ as)  []    = not (Set.null (Set.intersection ss as))
match (Regex n ss tf as) (i:is) = match (Regex n ss' tf as) is
  where
    ss' = Set.unions (map (tf i) (Set.elems ss))

-- ====================== SAMPLE CODE USING MATCH =================
-- regular expression that recognizes any string containing "Hello"
-- can build manually by representing NDFSM:
hello :: Regex Char
hello = Regex 6 (Set.singleton 0) f (Set.singleton 5)
  where
    f 'H' 0 = Set.fromList [0,1]
    f 'e' 1 = Set.fromList [0,2]
    f 'l' 2 = Set.fromList [0,3]
    f 'l' 3 = Set.fromList [0,4]
    f 'o' 4 = Set.fromList [0,5]
    f _ 5 = Set.singleton 5 -- stay in 5 forever as saw "Hello"
    f  _  _ = Set.singleton 0 -- if get something unexpected go to 0

-- Use: match hello "Hello" -> returns True

-- or can build using regular expression generators.  This recognizes
-- only "Hello", fails if anything before or after it.
rhello = cat
           (cat
              (cat
                 (cat (once 'H') (once 'e'))
                 (once 'l'))
              (once 'l'))
           (once 'o')

-- matches any string containing "Hello" like hello above
anyhelloany = cat (star dot) (cat rhello (star dot))

ac = once 'a'  -- recognize just a single 'a'
bc = once 'b'  -- recognize just a single 'b'

-- regular expression (a|b)
aorb = union ac bc

-- (a|b)*
aorbstar = star aorb

-- (a|b)* aa (a|b)*
containaa = cat (cat aorbstar (cat ac ac)) aorbstar